------------------------------------------------------------------------
--- Linearity optimizer for function patterns and non-strict unification
---
--- Replace occurrences of (fp =:<= x) by (fp =:<<= x) if the function
--- pattern fp always evaluates to a linear term.
---
--- @author Michael Hanus
--- @version January 2006
------------------------------------------------------------------------

import FlatCurry
import FlatCurryRead
import Linearity
import Dependency
import FileGoodies
import System
import List(intersperse)
import ReadShowTerm(showTerm)
import Directory(doesFileExist)
import IO

-- Example: optimizeNonstrictEqualityInModule "last"

-- Check arguments and call main function:
main = do
  args <- getArgs
  case args of
    [prog] -> optimizeNonstrictEqualityInModuleIfNecessary (stripSuffix prog)
    _ -> putStrLn $ "ERROR: Illegal arguments: " ++
                    concat (intersperse " " args) ++ "\n" ++
                    "Usage: OptNonStrict.state <module_name>"

optimizeNonstrictEqualityInModuleIfNecessary prog = do
  optimized <- hasCommentOption (prog++".fcy") "-fpopt"
  if optimized
   then putStrLn ("Program '"++prog++".fcy' already optimized")
   else optimizeNonstrictEqualityInModule prog

optimizeNonstrictEqualityInModule prog = do
  progs <- readFlatCurryWithImports prog
  let (nsu,lnsu,optmod) = optimizeNonstrictEquality (concatMap funcsOfProg progs)
                                                    (head progs)
  writeFile (prog++"_lo.fcy") ("{- -fpopt -}\n" ++ showTerm optmod)
  system ("mv "++prog++"_lo.fcy "++prog++".fcy")
  putStrLn (show nsu ++ " non-strict unifications found, " ++ show lnsu ++
            " optimized into linear unifications.")
  putStrLn ("Optimized program written to '"++prog++".fcy'")


------------------------------------------------------------------------------
-- Checks whether a file exists and the file starts with a comment of the form
--  "{- ...-}" and contains the second argument as a word in the comment:
hasCommentOption :: String -> String -> IO Bool
hasCommentOption filename option = do
  existsfile <- doesFileExist filename
  if existsfile
   then do cwords <- readWordsInFirstCommentLine filename
           return (option `elem` cwords)
   else return False

-- Read the words in the first comment line and return them, if the file
-- start with a comment, otherwise return the empty list:
readWordsInFirstCommentLine :: String -> IO [String]
readWordsInFirstCommentLine filename = do
  fh <- openFile filename ReadMode
  c1 <- hGetChar fh
  c2 <- hGetChar fh
  if c1=='{' && c2=='-'
   then do cs <- getCommentString fh
           hClose fh
           return (words cs)
   else do hClose fh
           return []

getCommentString fh = do
  c <- hGetChar fh
  if c=='-' then getCommentString' fh
            else do cs <- getCommentString fh
                    return (c:cs)

getCommentString' fh = do
  c <- hGetChar fh
  if c=='}' then return []
            else if c=='-'
                 then do cs <- getCommentString' fh
                         return ('-':cs)
                 else do cs <- getCommentString fh
                         return ('-':c:cs)


------------------------------------------------------------------------------
-- Replace "=:<=" by "=:<<=" if the left argument is a linear term with
-- linear functions.
-- The first argument is the list of all functions occuring in the program,
-- the second argument is the current module to be optimized.
-- The result is a tuple containing the number of "=:<=" occurring in the
-- module, the number of the optimized occurrences, and the optimized module.
optimizeNonstrictEquality :: [FuncDecl] -> Prog -> (Int,Int,Prog)
optimizeNonstrictEquality allfuns (Prog mod imps ts funs ops) =
  let (nsus,lnsus,optfs) = unzip3 (map (optimizeFun (analyseRightLinearity allfuns)) funs)
  in (sum nsus, sum lnsus, Prog mod imps ts optfs ops)

optimizeFun lininfo (Func qn ar vis ty (Rule vs e)) =
  let (nsu,lnsu,opte) = optimizeExp lininfo e
  in (nsu,lnsu,Func qn ar vis ty (Rule vs opte))
optimizeFun _ (Func qn ar vis ty (External e)) =
  (0,0,Func qn ar vis ty (External e))

optimizeExp :: [(QName,Bool)] -> Expr -> (Int,Int,Expr)
optimizeExp _ (Var i) = (0,0,Var i)
optimizeExp _ (Lit l) = (0,0,Lit l)
optimizeExp lininfo (Comb ct f es)
 | f==("prelude","=:<=") && length es == 2
  = let e1 = head es
        e2 = head (tail es)
        (nsu1,lnsu1,opte1) = optimizeExp lininfo e1
        (nsu2,lnsu2,opte2) = optimizeExp lininfo e2
    in if linearExpr e1 && onlyLinearFunctions lininfo e1
       then (nsu1+nsu2+1,lnsu1+lnsu2+1,Comb ct ("prelude","=:<<=") [opte1,opte2])
       else (nsu1+nsu2+1,lnsu1+lnsu2,Comb ct ("prelude","=:<=") [opte1,opte2])
 | otherwise
  = let (nsus,lnsus,optes) = unzip3 (map (optimizeExp lininfo) es)
    in (sum nsus, sum lnsus, Comb ct f optes)
optimizeExp lininfo (Free vs e) = 
  let (nsu,lnsu,opte) = optimizeExp lininfo e
  in  (nsu,lnsu,Free vs opte)
optimizeExp li (Let bs exp) =
  let (nsu,lnsu,optexp) = optimizeExp li exp
      (bvs,bes) = unzip bs
      (nsus,lnsus,optbes) = unzip3 (map (optimizeExp li) bes)
  in (nsu + sum nsus, lnsu + sum lnsus, Let (zip bvs optbes) optexp)
optimizeExp lininfo (Or e1 e2) =
  let (nsu1,lnsu1,opte1) = optimizeExp lininfo e1
      (nsu2,lnsu2,opte2) = optimizeExp lininfo e2
  in (nsu1+nsu2, lnsu1+lnsu2, Or opte1 opte2)
optimizeExp li (Case ct exp bs) =
  let (nsu,lnsu,optexp) = optimizeExp li exp
      (nsus,lnsus,optbs) = unzip3 (map optimizeBranch bs)
  in (nsu + sum nsus, lnsu + sum lnsus, Case ct optexp optbs)
 where
  optimizeBranch (Branch patt be) =
    let (nsub,lnsub,optbe) = optimizeExp li be
    in (nsub,lnsub,Branch patt optbe)

onlyLinearFunctions :: [(QName,Bool)] -> Expr -> Bool
onlyLinearFunctions li e = all isRightLinearDefined (funcsInExpr e)
 where
   isRightLinearDefined fun = maybe False id (lookup fun li)

-- goodies:

funcsOfProg (Prog _ _ _ funcs _) = funcs

sum = foldl (+) 0
